
#ifndef H_BOARD
 #define H_BOARD

#include <stdio.h>
#include <string>
#include <vector>

typedef char Player;
#define PLAYER_X 'X'
#define PLAYER_O 'O'
#define NOBODY   ' '

enum Turn {
    X_TO_MOVE,
    O_TO_COLLAPSE,
    O_TO_MOVE,
    X_TO_COLLAPSE
};

struct Move {
    Turn kind;
    /* if kind is TO_MOVE */
    int firstcell;
    int secondcell;
    /* if kind is TO_COLLAPSE */
    bool collapse_into_first;
    
    bool is_to_move() const { return (kind==X_TO_MOVE || kind==O_TO_MOVE); }
    bool is_to_collapse() const { return !is_to_move(); }
    std::string toString() const;
};

struct WhoWhen {
    Player who;
    int when;
    WhoWhen() { }
    WhoWhen(Player wo, int we): who(wo), when(we) { }
    bool operator == (const WhoWhen &rhs) const
    { return (who == rhs.who) && (who == NOBODY || when == rhs.when); }
};

struct Board {
    /* If the turn is X_TO_COLLAPSE, then last_move will be relevant,
     * and the attacker's only choices will be to collapse
     * the defender's play into last_move.firstcell or last_move.secondcell.
     * The rest of the cycle will collapse accordingly. */
    Turn turn;
    WhoWhen collapsed_cells[9];
    WhoWhen adj[9][9];
    
    int now;
    /* if turn is TO_COLLAPSE */
    Move last_move;

    void new_game();
    
    bool cell_is_collapsed(int cell) const;
    bool is_valid_move(const Move &m) const;
    bool path_exists_between(int a, int b) const;
    Player attacker() const { return (turn==X_TO_MOVE || turn==X_TO_COLLAPSE) ? 'X' : 'O'; }
    Player defender() const { return (turn==X_TO_MOVE || turn==X_TO_COLLAPSE) ? 'O' : 'X'; }
    bool is_to_move() const { return (turn==X_TO_MOVE || turn==O_TO_MOVE); }
    bool is_to_collapse() const { return !is_to_move(); }
    Player winner() const;
    bool game_over() const;
    void print(FILE *out) const;
};


void applymove(Board &b, const Move &move);
void findmoves(const Board &b, std::vector<Move> &allmoves);
int findattacker(const Board &b);

#endif /* H_BOARD */
