
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <vector>
#include "Board.h"
#include "AlphaBeta.h"


Move get_human_move(const Board &b)
{
    static char buffer[1000];
    Move move;
    move.kind = b.turn;

    b.print(stdout);
    puts("Enter your move.");
    if (b.is_to_move())
      puts("e.g. to link cells 1 and 7, type \"17\".");
    else
      puts("To collapse the given edge to its first cell, type \"y\"; else type \"n\".");
  get_move:
    gets(buffer);
    if (b.is_to_move()) {
        if ('1' <= buffer[0] && buffer[0] <= '9')
          move.firstcell = (buffer[0] - '1');
        else goto get_move_again;
        if ('1' <= buffer[1] && buffer[1] <= '9')
          move.secondcell = (buffer[1] - '1');
        else goto get_move_again;
    } else {
        if (buffer[0] == 'y')
          move.collapse_into_first = true;
        else if (buffer[0] == 'n')
          move.collapse_into_first = false;
        else goto get_move_again;
    }
    if (!b.is_valid_move(move)) goto get_move_again;
    return move;
  get_move_again:
    puts("I didn't understand that. Please try again.");
    goto get_move;
}


static void findmoves_randomorder(const Board &b, std::vector<Move> &allmoves)
{
    assert(allmoves.empty());
    findmoves(b, allmoves);
    std::random_shuffle(allmoves.begin(), allmoves.end());
}

/* Return the value of this board to the defender. */
int evaluate(const Board &b)
{
    const Player winner = b.winner();
    const Player defender = b.defender();
    if (winner != NOBODY) {
        return (winner == b.defender()) ? +500 : -500;
    } else {
        int value = 0;
        /* Having a mark in the center or in a corner is valuable. */
        /* Having a collapsed mark in those spots is very valuable. */
        for (int i=0; i < 9; ++i) {
#define MULT(who) (who==defender ? +1 : who==NOBODY ? 0 : -1)
#define VAL(cell) (cell==5 ? +2 : (cell==1 || cell==3 || cell==7 || cell==9) ? +1 : 0)
            value += (VAL(i) + 2) * MULT(b.collapsed_cells[i].who);
            for (int j=0; j < i; ++j) {
                value += VAL(i) * MULT(b.adj[i][j].who);
            }
        }
        return value;
    }
}

Move get_alphabeta_move(const Board &b)
{
    static AlphaBeta<Board,Move,int> ab(evaluate, applymove,
                                        findmoves_randomorder, findattacker);

    if (b.now == 1) {
        /* The first move should always be this. */
        assert(b.turn == X_TO_MOVE);
        Move firstmove;
        firstmove.kind = X_TO_MOVE;
        firstmove.firstcell = 0;
        firstmove.secondcell = 4;
        return firstmove;
    }

    Move bestmove;
    int bestvalue;
#if 0
    const int ply = (b.now < 4) ? 4 : 8;
    ab.depth_first_alpha_beta(b, ply, bestmove, bestvalue, -1000, +1000);
#else
    ab.breadth_first(b, 200000, bestmove, bestvalue);
#endif
    printf("Computer thinks bestvalue is %d\n", bestvalue);
    return bestmove;
}

int main(int argc, char **argv)
{
    srand((unsigned int)time(NULL));
    Board b;
    b.new_game();
    Player computer = 'X';
    if (argc >= 2 && argv[1][0] == 'X') computer = 'O';
    while (!b.game_over()) {
        Move move;
        if (b.attacker() == computer) {
            move = get_alphabeta_move(b);
            assert(b.is_valid_move(move));
            printf("Computer made this move: %s\n", move.toString().c_str());
        } else {
            move = get_human_move(b);
            assert(b.is_valid_move(move));
        }
        applymove(b, move);
    }
    const Player winner = b.winner();
    if (winner == computer) {
        b.print(stdout);
        puts("You lost to the computer!");
    } else if (winner == NOBODY) {
        b.print(stdout);
        puts("Cat game.");
    } else {
        b.print(stdout);
        puts("Human beat the computer!");
    }
    return 0;
}
